#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Table
# Table management.
# Specification: workspace:v1.0
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Table(Namespace):

    scope = "workspace"

    def createTable(self, table_name: str, db_name: str = None, branch_name: str = None) -> Response:
        """
        Creates a new table with the given name.  Returns 422 if a table with the same name
        already exists.

        Path: /db/{db_branch_name}/tables/{table_name}
        Method: PUT
        Response status codes:
        - 201: Created
        - 204: No Content
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 422: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error
        Response: application/json

        :param table_name: str The Table name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}"
        return self.request("PUT", url_path)

    def deleteTable(self, table_name: str, db_name: str = None, branch_name: str = None) -> Response:
        """
        Deletes the table with the given name.

        Path: /db/{db_branch_name}/tables/{table_name}
        Method: DELETE
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Not Found
        - 5XX: Unexpected Error
        - default: Unexpected Error
        Response: application/json

        :param table_name: str The Table name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}"
        return self.request("DELETE", url_path)

    def updateTable(self, table_name: str, payload: dict, db_name: str = None, branch_name: str = None) -> Response:
        """
        Update table.  Currently there is only one update operation supported: renaming the table
        by providing a new name.  In the example below, we rename a table from “users” to
        “people”:  ```json // PATCH /db/test:main/tables/users  {   "name": "people" } ```

        Path: /db/{db_branch_name}/tables/{table_name}
        Method: PATCH
        Response status codes:
        - 200: Schema migration response with ID and migration status.
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 422: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error

        :param table_name: str The Table name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}"
        headers = {"content-type": "application/json"}
        return self.request("PATCH", url_path, headers, payload)

    def getTableSchema(self, table_name: str, db_name: str = None, branch_name: str = None) -> Response:
        """
        Get table schema

        Path: /db/{db_branch_name}/tables/{table_name}/schema
        Method: GET
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error
        Response: application/json

        :param table_name: str The Table name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/schema"
        return self.request("GET", url_path)

    def setTableSchema(self, table_name: str, payload: dict, db_name: str = None, branch_name: str = None) -> Response:
        """
        Update table schema

        Path: /db/{db_branch_name}/tables/{table_name}/schema
        Method: PUT
        Response status codes:
        - 200: Schema migration response with ID and migration status.
        - 204: No Content
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 409: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error

        :param table_name: str The Table name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/schema"
        headers = {"content-type": "application/json"}
        return self.request("PUT", url_path, headers, payload)

    def getTableColumns(self, table_name: str, db_name: str = None, branch_name: str = None) -> Response:
        """
        Retrieves the list of table columns and their definition.  This endpoint returns the
        column list with object columns being reported with their full dot-separated path
        (flattened).

        Path: /db/{db_branch_name}/tables/{table_name}/columns
        Method: GET
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error
        Response: application/json

        :param table_name: str The Table name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/columns"
        return self.request("GET", url_path)

    def addTableColumn(self, table_name: str, payload: dict, db_name: str = None, branch_name: str = None) -> Response:
        """
        Adds a new column to the table.  The body of the request should contain the column
        definition.

        Path: /db/{db_branch_name}/tables/{table_name}/columns
        Method: POST
        Response status codes:
        - 200: Schema migration response with ID and migration status.
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error

        :param table_name: str The Table name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/columns"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def getColumn(self, table_name: str, column_name: str, db_name: str = None, branch_name: str = None) -> Response:
        """
        Get the definition of a single column.

        Path: /db/{db_branch_name}/tables/{table_name}/columns/{column_name}
        Method: GET
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error
        Response: application/json

        :param table_name: str The Table name
        :param column_name: str The Column name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/columns/{column_name}"
        return self.request("GET", url_path)

    def deleteColumn(self, table_name: str, column_name: str, db_name: str = None, branch_name: str = None) -> Response:
        """
        Deletes the specified column.

        Path: /db/{db_branch_name}/tables/{table_name}/columns/{column_name}
        Method: DELETE
        Response status codes:
        - 200: Schema migration response with ID and migration status.
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error

        :param table_name: str The Table name
        :param column_name: str The Column name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/columns/{column_name}"
        return self.request("DELETE", url_path)

    def updateColumn(
        self, table_name: str, column_name: str, payload: dict, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Update column with partial data.  Can be used for renaming the column by providing a new
        "name" field.

        Path: /db/{db_branch_name}/tables/{table_name}/columns/{column_name}
        Method: PATCH
        Response status codes:
        - 200: Schema migration response with ID and migration status.
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error

        :param table_name: str The Table name
        :param column_name: str The Column name
        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/columns/{column_name}"
        headers = {"content-type": "application/json"}
        return self.request("PATCH", url_path, headers, payload)



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from .client import XataClient

__all__ = ("XataClient", "BulkProcessor", "to_rfc3339", "Transaction")



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Workspaces
# Workspaces management
# Specification: core:v1.0
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Workspaces(Namespace):

    scope = "core"

    def getWorkspacesList(
        self,
    ) -> Response:
        """
        Retrieve the list of workspaces the user belongs to

        Path: /workspaces
        Method: GET
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        Response: application/json


        :return Response
        """
        url_path = "/workspaces"
        return self.request("GET", url_path)

    def createWorkspace(self, payload: dict) -> Response:
        """
        Creates a new workspace with the user requesting it as its single owner.

        Path: /workspaces
        Method: POST
        Response status codes:
        - 201: Created
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        Response: application/json

        :param payload: dict content

        :return Response
        """
        url_path = "/workspaces"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def getWorkspace(self, workspace_id: str = None) -> Response:
        """
        Retrieve workspace info from a workspace ID

        Path: /workspaces/{workspace_id}
        Method: GET
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 403: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        Response: application/json

        :param workspace_id: str = None The workspace identifier. Default: workspace Id from the client.

        :return Response
        """
        if workspace_id is None:
            workspace_id = self.client.get_config()["workspaceId"]
        url_path = f"/workspaces/{workspace_id}"
        return self.request("GET", url_path)

    def updateWorkspace(self, payload: dict, workspace_id: str = None) -> Response:
        """
        Update workspace info

        Path: /workspaces/{workspace_id}
        Method: PUT
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 403: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        Response: application/json

        :param payload: dict content
        :param workspace_id: str = None The workspace identifier. Default: workspace Id from the client.

        :return Response
        """
        if workspace_id is None:
            workspace_id = self.client.get_config()["workspaceId"]
        url_path = f"/workspaces/{workspace_id}"
        headers = {"content-type": "application/json"}
        return self.request("PUT", url_path, headers, payload)

    def deleteWorkspace(self, workspace_id: str = None) -> Response:
        """
        Delete the workspace with the provided ID

        Path: /workspaces/{workspace_id}
        Method: DELETE
        Response status codes:
        - 204: No Content
        - 400: Bad Request
        - 401: Authentication Error
        - 403: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error

        :param workspace_id: str = None The workspace identifier. Default: workspace Id from the client.

        :return Response
        """
        if workspace_id is None:
            workspace_id = self.client.get_config()["workspaceId"]
        url_path = f"/workspaces/{workspace_id}"
        return self.request("DELETE", url_path)

    def getWorkspaceMembersList(self, workspace_id: str = None) -> Response:
        """
        Retrieve the list of members of the given workspace

        Path: /workspaces/{workspace_id}/members
        Method: GET
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 403: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        Response: application/json

        :param workspace_id: str = None The workspace identifier. Default: workspace Id from the client.

        :return Response
        """
        if workspace_id is None:
            workspace_id = self.client.get_config()["workspaceId"]
        url_path = f"/workspaces/{workspace_id}/members"
        return self.request("GET", url_path)

    def updateWorkspaceMemberRole(self, user_id: str, payload: dict, workspace_id: str = None) -> Response:
        """
        Update a workspace member role.  Workspaces must always have at least one owner, so this
        operation will fail if trying to remove owner role from the last owner in the workspace.

        Path: /workspaces/{workspace_id}/members/{user_id}
        Method: PUT
        Response status codes:
        - 204: No Content
        - 400: Bad Request
        - 401: Authentication Error
        - 403: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error

        :param user_id: str UserID
        :param payload: dict content
        :param workspace_id: str = None The workspace identifier. Default: workspace Id from the client.

        :return Response
        """
        if workspace_id is None:
            workspace_id = self.client.get_config()["workspaceId"]
        url_path = f"/workspaces/{workspace_id}/members/{user_id}"
        headers = {"content-type": "application/json"}
        return self.request("PUT", url_path, headers, payload)

    def removeWorkspaceMember(self, user_id: str, workspace_id: str = None) -> Response:
        """
        Remove the member from the workspace

        Path: /workspaces/{workspace_id}/members/{user_id}
        Method: DELETE
        Response status codes:
        - 204: No Content
        - 400: Bad Request
        - 401: Authentication Error
        - 403: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error

        :param user_id: str UserID
        :param workspace_id: str = None The workspace identifier. Default: workspace Id from the client.

        :return Response
        """
        if workspace_id is None:
            workspace_id = self.client.get_config()["workspaceId"]
        url_path = f"/workspaces/{workspace_id}/members/{user_id}"
        return self.request("DELETE", url_path)



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Invites
# Manage user invites.
# Specification: core:v1.0
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Invites(Namespace):

    scope = "core"

    def inviteWorkspaceMember(self, payload: dict, workspace_id: str = None) -> Response:
        """
        Invite some user to join the workspace with the given role

        Path: /workspaces/{workspace_id}/invites
        Method: POST
        Response status codes:
        - 201: Created
        - 400: Bad Request
        - 401: Authentication Error
        - 403: Authentication Error
        - 404: Example response
        - 409: Example response
        - 5XX: Unexpected Error
        Response: application/json

        :param payload: dict content
        :param workspace_id: str = None The workspace identifier. Default: workspace Id from the client.

        :return Response
        """
        if workspace_id is None:
            workspace_id = self.client.get_config()["workspaceId"]
        url_path = f"/workspaces/{workspace_id}/invites"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def cancelWorkspaceMemberInvite(self, invite_id: str, workspace_id: str = None) -> Response:
        """
        This operation provides a way to cancel invites by deleting them.  Already accepted
        invites cannot be deleted.

        Path: /workspaces/{workspace_id}/invites/{invite_id}
        Method: DELETE
        Response status codes:
        - 204: No Content
        - 400: Bad Request
        - 401: Authentication Error
        - 403: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error

        :param invite_id: str Invite identifier
        :param workspace_id: str = None The workspace identifier. Default: workspace Id from the client.

        :return Response
        """
        if workspace_id is None:
            workspace_id = self.client.get_config()["workspaceId"]
        url_path = f"/workspaces/{workspace_id}/invites/{invite_id}"
        return self.request("DELETE", url_path)

    def updateWorkspaceMemberInvite(self, invite_id: str, payload: dict, workspace_id: str = None) -> Response:
        """
        This operation provides a way to update an existing invite.  Updates are performed in-
        place; they do not change the invite link, the expiry time, nor do they re-notify the
        recipient of the invite.

        Path: /workspaces/{workspace_id}/invites/{invite_id}
        Method: PATCH
        Response status codes:
        - 200: Updated successfully.
        - 400: Bad Request
        - 401: Authentication Error
        - 403: Authentication Error
        - 404: Example response
        - 422: Example response
        - 5XX: Unexpected Error
        Response: application/json

        :param invite_id: str Invite identifier
        :param payload: dict content
        :param workspace_id: str = None The workspace identifier. Default: workspace Id from the client.

        :return Response
        """
        if workspace_id is None:
            workspace_id = self.client.get_config()["workspaceId"]
        url_path = f"/workspaces/{workspace_id}/invites/{invite_id}"
        headers = {"content-type": "application/json"}
        return self.request("PATCH", url_path, headers, payload)

    def acceptWorkspaceMemberInvite(self, invite_key: str, workspace_id: str = None) -> Response:
        """
        Accept the invitation to join a workspace.  If the operation succeeds the user will be a
        member of the workspace

        Path: /workspaces/{workspace_id}/invites/{invite_key}/accept
        Method: POST
        Response status codes:
        - 204: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 403: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error

        :param invite_key: str Invite Key (secret) for the invited user
        :param workspace_id: str = None The workspace identifier. Default: workspace Id from the client.

        :return Response
        """
        if workspace_id is None:
            workspace_id = self.client.get_config()["workspaceId"]
        url_path = f"/workspaces/{workspace_id}/invites/{invite_key}/accept"
        return self.request("POST", url_path)

    def resendWorkspaceMemberInvite(self, invite_id: str, workspace_id: str = None) -> Response:
        """
        This operation provides a way to resend an Invite notification.  Invite notifications can
        only be sent for Invites not yet accepted.

        Path: /workspaces/{workspace_id}/invites/{invite_id}/resend
        Method: POST
        Response status codes:
        - 204: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 403: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error

        :param invite_id: str Invite identifier
        :param workspace_id: str = None The workspace identifier. Default: workspace Id from the client.

        :return Response
        """
        if workspace_id is None:
            workspace_id = self.client.get_config()["workspaceId"]
        url_path = f"/workspaces/{workspace_id}/invites/{invite_id}/resend"
        return self.request("POST", url_path)



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import logging
import time
from datetime import datetime, timezone
from threading import Lock, Thread

from .client import XataClient

BP_DEFAULT_THREAD_POOL_SIZE = 4
BP_DEFAULT_BATCH_SIZE = 25
BP_DEFAULT_FLUSH_INTERVAL = 5
BP_DEFAULT_PROCESSING_TIMEOUT = 0.025
BP_DEFAULT_THROW_EXCEPTION = False
BP_VERSION = "0.2.1"
TRX_MAX_OPERATIONS = 1000
TRX_VERSION = "0.0.1"


class BulkProcessor(object):
    """
    Additional abstraction for bulk requests that process'
    requests in parallel
    :stability beta
    """

    def __init__(
        self,
        client: XataClient,
        thread_pool_size: int = BP_DEFAULT_THREAD_POOL_SIZE,
        batch_size: int = BP_DEFAULT_BATCH_SIZE,
        flush_interval: int = BP_DEFAULT_FLUSH_INTERVAL,
        processing_timeout: float = BP_DEFAULT_PROCESSING_TIMEOUT,
        throw_exception: bool = BP_DEFAULT_THROW_EXCEPTION,
    ):
        """
        BulkProcessor: Abstraction for bulk ingestion of records.

        :stability beta

        :param client: XataClient
        :param thread_pool_size: int How many data queue workers should be deployed (default: 4)
        :param batch_size: int How many records per table should be pushed as batch (default: 25)
        :param flush_interval: int After how many seconds should the per table queue be flushed (default: 5 seconds)
        :processing_timeout: float Cooldown period between batches (default: 0.025 seconds)
        :throw_exception: bool Throw exception ingestion, could kill all workers (default: False)

        :raises Exception if throw exception is enabled
        """
        if thread_pool_size < 1:
            raise Exception("thread pool size must be greater than 0, default: %d" % BP_DEFAULT_THREAD_POOL_SIZE)
        if processing_timeout < 0:
            raise Exception("processing timeout can not be negative, default: %f" % BP_DEFAULT_PROCESSING_TIMEOUT)
        if flush_interval < 0:
            raise Exception("flush interval can not be negative, default: %f" % BP_DEFAULT_FLUSH_INTERVAL)
        if batch_size < 1:
            raise Exception("batch size can not be less than one, default: %d" % BP_DEFAULT_BATCH_SIZE)

        self.client = client
        self.client.set_header("x-xata-helper", f"bulkprocessor/{BP_VERSION}")
        self.processing_timeout = processing_timeout
        self.batch_size = batch_size
        self.flush_interval = flush_interval
        self.failed_batches_queue = []
        self.throw_exception = throw_exception
        self.stats = {"total": 0, "queue": 0, "failed_batches": 0, "tables": {}}
        self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")

        self.thread_workers = []
        self.records = self.Records(self.batch_size, self.flush_interval, self.logger)

        for i in range(thread_pool_size):
            worker = Thread(target=self.process, daemon=True, args=(i,), name="worker-%d" % i)
            worker.start()
            self.thread_workers.append(worker)

    def process(self, id: int):
        """
        Process the records
        """
        self.logger.debug(
            "thread #%d: starting bulk processor [thread_pool=%d, batch=%d, flush=%d, processing_timeout=%s"
            % (
                id,
                len(self.thread_workers),
                self.batch_size,
                self.flush_interval,
                self.processing_timeout,
            )
        )
        while True:
            batch = self.records.next_batch()
            if "table" in batch and len(batch["records"]) > 0:
                try:
                    r = self.client.records().bulkInsertTableRecords(batch["table"], {"records": batch["records"]})
                    if r.status_code != 200:
                        self.logger.error(
                            "thread #%d: unable to process batch for table '%s', with error: %d - %s"
                            % (id, batch["table"], r.status_code, r.json())
                        )
                        # Add to failed queue
                        self.failed_batches_queue.append(
                            {
                                "timestamp": datetime.utcnow(),
                                "records": batch["records"],
                                "table": batch["table"],
                                "response": r,
                            }
                        )
                        self.stats["failed_batches"] += 1
                        if self.throw_exception:
                            raise Exception(r.json())

                    self.logger.debug(
                        "thread #%d: pushed a batch of %d records to table %s"
                        % (id, len(batch["records"]), batch["table"])
                    )
                    self.stats["total"] += len(batch["records"])
                    self.stats["queue"] = self.records.size()
                    if batch["table"] not in self.stats["tables"]:
                        self.stats["tables"][batch["table"]] = 0
                    self.stats["tables"][batch["table"]] += len(batch["records"])
                except Exception as exc:
                    logging.error("thread #%d: %s" % (id, exc))
            time.sleep(self.processing_timeout)

    def put_record(self, table_name: str, record: dict):
        """
        Put a record to the processing queue

        :param table_name: str
        :param record: dict
        """
        self.records.put(table_name, [record])

    def put_records(self, table_name: str, records: list[dict]):
        """
        Put multtiple records to the processing queue

        :param table_name: str
        :param records: list[dict]
        """
        self.records.put(table_name, records)

    def get_failed_batches(self) -> list[dict]:
        """
        Get the batched records that could not be processed with the error
        :return list[dict]
        """
        return self.failed_batches_queue

    def get_stats(self):
        """
        Get processing statistics

        :return dict
        """
        return self.stats

    def flush_queue(self):
        """
        Flush all records from the queue.
        """
        self.logger.debug("flushing queue with %d records .." % (self.records.size()))
        self.records.set_flush_interval(0)
        self.processing_timeout = 0

        # If the queue is not empty wait for one flush interval.
        # Purpose is a race condition with self.stats["queue"]
        if self.records.size() > 0:
            time.sleep(self.flush_interval)

        while self.stats["queue"] > 0:
            self.logger.debug("flushing queue with %d records." % self.stats["queue"])
            time.sleep(self.processing_timeout / len(self.thread_workers) + 0.01)

    class Records(object):
        """
        Thread safe storage for records to persist by the bulk processor
        """

        def __init__(self, batch_size: int, flush_interval: int, logger):
            """
            :param batch_size: int
            :param flush_interval: int
            """
            self.batch_size = batch_size
            self.flush_interval = flush_interval
            self.logger = logger

            self.store = dict()
            self.store_ptr = 0
            self.lock = Lock()

        def set_flush_interval(self, interval: int):
            self.flush_interval = interval

        def put(self, table_name: str, records: list[dict]):
            """
            :param table_name: str
            :param records: list[dict]
            """
            with self.lock:
                if table_name not in self.store.keys():
                    self.store[table_name] = {
                        "lock": Lock(),
                        "flushed": time.time(),
                        "records": list(),
                    }
            with self.store[table_name]["lock"]:
                self.store[table_name]["records"] += records

        def next_batch(self) -> dict:
            """
            Get the next batch of records to persist

            :return dict
            """
            table_name = ""
            with self.lock:
                names = list(self.store.keys())
                if len(names) == 0:
                    return {}

                self.store_ptr += 1
                if len(names) <= self.store_ptr:
                    self.store_ptr = 0
                table_name = names[self.store_ptr]

            rs = []
            with self.store[table_name]["lock"]:
                # flush interval exceeded
                time_elapsed = time.time() - self.store[table_name]["flushed"]
                flush_needed = time_elapsed > self.flush_interval
                if flush_needed and len(self.store[table_name]["records"]) > 0:
                    self.logger.debug(
                        "flushing table '%s' with %d records after interval %s > %d"
                        % (
                            table_name,
                            len(self.store[table_name]["records"]),
                            time_elapsed,
                            self.flush_interval,
                        )
                    )
                # pop records ?
                if len(self.store[table_name]["records"]) >= self.batch_size or flush_needed:
                    self.store[table_name]["flushed"] = time.time()
                    rs = self.store[table_name]["records"][0 : self.batch_size]
                    del self.store[table_name]["records"][0 : self.batch_size]
            return {"table": table_name, "records": rs}

        def length(self, table_name: str) -> int:
            """
            Get record count of a table

            :param table_name: str
            """
            with self.store[table_name]["lock"]:
                return len(self.store[table_name]["records"])

        def size(self) -> int:
            """
            Get total size of stored records
            """
            with self.lock:
                return sum([len(self.store[n]["records"]) for n in self.store.keys()])


def to_rfc339(dt: datetime, tz=timezone.utc) -> str:
    """
    Format a datetime object to an RFC3339 compliant string
    :link https://xata.io/docs/concepts/data-model#datetime
    :link https://datatracker.ietf.org/doc/html/rfc3339

    :param dt: datetime instance to convert
    :param tz: timezone to convert in, default: UTC
    :return str
    """
    return dt.replace(tzinfo=tz).isoformat()


class Transaction(object):
    """
    Additional abstraction for bulk requests that process'
    requests in parallel
    :stability beta
    """

    def __init__(
        self,
        client: XataClient,
    ) -> None:
        """
        Transaction Helper
        Wrapper to simplify running transactions

        :stability beta

        :param client: XataClient
        """
        self.client = client
        self.client.set_header("x-xata-helper", f"transaction/{TRX_VERSION}")

        self.has_run = False
        self.operations = {"operations": []}

    def _add_operation(self, operation: dict) -> None:
        if len(self.operations["operations"]) >= TRX_MAX_OPERATIONS:
            raise Exception(f"Maximum amount of {TRX_MAX_OPERATIONS} transaction operations exceeded.")
        self.operations["operations"].append(operation)

    def insert(self, table: str, record: dict, createOnly: bool = False) -> None:
        """
        Inserts can be used to insert records across any number of tables in your database. As with the
        insert endpoints, you can explicitly set an ID, or omit it and have Xata auto-generate one for you.
        Either way, on a successful transaction, Xata will return the ID to you.

        :param table: str
        :param record: dict
        :param createOnly: bool By default, if a record exists with the same explicit ID, Xata will overwrite
            the record. You can adjust this behavior by setting `createOnly` to `true` for the operation. Default: False

        :raises Exception if limit of 1000 operations is exceeded
        """
        self._add_operation({"insert": {"table": table, "record": record, "createOnly": createOnly}})

    def update(self, table: str, recordId: str, fields: dict, upsert: bool = False) -> None:
        """
        Updates can be used to update records in any number of tables in your database. The update operation
        requires an ID parameter explicitly defined. The operation will only replace the fields explicitly
        specified in your operation. The update operation also supports the upsert flag. Off by default, but
        if set to `true`, the update operation will insert the record if no record is found with the provided ID.

        :param table: str
        :param recordId: str
        :param fields: dict
        :param upsert: bool Default: False

        :raises Exception if limit of 1000 operations is exceeded
        """
        self._add_operation({"update": {"table": table, "id": recordId, "fields": fields, "upsert": upsert}})

    def delete(self, table: str, recordId: str, columns: list[str] = []) -> None:
        """
        A delete is used to remove records. Delete can operate on records from the same transaction, and will
        not cancel a transaction if no record is found.

        :param table: str
        :param recordId: str
        :param columns: list of columns to retrieve

        :raises Exception if limit of 1000 operations is exceeded
        """
        self._add_operation({"delete": {"table": table, "id": recordId, "columns": columns}})

    def get(self, table: str, recordId: str, columns: list[str] = []) -> None:
        """
        A get is used to retrieve a record by id. A get operation can retrieve records created in the same transaction
        but will not cancel a transaction if no record is found.

        :param table: str
        :param recordId: str
        :param columns: list of columns to retrieve

        :raises Exception if limit of 1000 operations is exceeded
        """
        self._add_operation({"get": {"table": table, "id": recordId, "columns": columns}})

    def run(self) -> dict:
        """
        Commit the transactions. Flushes the operations queue

        :return dict
        """
        r = self.client.records().branchTransaction(self.operations)
        result = {
            "status_code": r.status_code,
            "results": r.json()["results"] if "results" in r.json() else [],
            "has_errors": True if "errors" in r.json() else False,
            "errors": r.json()["errors"] if "errors" in r.json() else [],
        }
        self.operations["operations"] = []  # free memory
        return result

    def size(self) -> int:
        """
        Get amount of operations in queue
        :return int
        """
        return len(self.operations["operations"])



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Files
# Files
# Specification: core:v1.0
# ------------------------------------------------------- #

from requests import Response, request

from xata.errors import ServerErrorException
from xata.namespace import Namespace


class Files(Namespace):

    scope = "workspace"

    def get_item(
        self,
        table_name: str,
        record_id: str,
        column_name: str,
        file_id: str,
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        Retrieves file content from an array by file ID

        Path: /db/{db_branch_name}/tables/{table_name}/data/{record_id}/column/{column_name}/file/{file_id}
        Method: GET
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        Response: */*

        :param table_name: str The Table name
        :param record_id: str The Record name
        :param column_name: str The Column name
        :param file_id: str The File Identifier
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/data/{record_id}/column/{column_name}/file/{file_id}"
        return self.request("GET", url_path)

    def put_item(
        self,
        table_name: str,
        record_id: str,
        column_name: str,
        file_id: str,
        data: bytes,
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        Uploads the file content to an array given the file ID

        Path: /db/{db_branch_name}/tables/{table_name}/data/{record_id}/column/{column_name}/file/{file_id}
        Method: PUT
        Response status codes:
        - 200: OK
        - 201: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 422: Example response
        - 5XX: Unexpected Error

        :param table_name: str The Table name
        :param record_id: str The Record name
        :param column_name: str The Column name
        :param file_id: str The File Identifier
        :param data: bytes content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/data/{record_id}/column/{column_name}/file/{file_id}"
        headers = {"content-type": "application/json"}
        return self.request("PUT", url_path, headers, data=data)

    def delete_item(
        self,
        table_name: str,
        record_id: str,
        column_name: str,
        file_id: str,
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        Deletes an item from an file array column given the file ID

        Path: /db/{db_branch_name}/tables/{table_name}/data/{record_id}/column/{column_name}/file/{file_id}
        Method: DELETE
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error

        :param table_name: str The Table name
        :param record_id: str The Record name
        :param column_name: str The Column name
        :param file_id: str The File Identifier
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/data/{record_id}/column/{column_name}/file/{file_id}"
        return self.request("DELETE", url_path)

    def get(
        self, table_name: str, record_id: str, column_name: str, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Retrieves the file content from a file column

        Path: /db/{db_branch_name}/tables/{table_name}/data/{record_id}/column/{column_name}/file
        Method: GET
        Response status codes:
        - 200: OK
        - 204: no content
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        Response: */*

        :param table_name: str The Table name
        :param record_id: str The Record name
        :param column_name: str The Column name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/data/{record_id}/column/{column_name}/file"
        return self.request("GET", url_path)

    def put(
        self,
        table_name: str,
        record_id: str,
        column_name: str,
        data: bytes,
        content_type: str = "application/octet-stream",
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        Uploads the file content to the given file column

        Path: /db/{db_branch_name}/tables/{table_name}/data/{record_id}/column/{column_name}/file
        Method: PUT
        Response status codes:
        - 200: OK
        - 201: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 422: Example response
        - 5XX: Unexpected Error

        :param table_name: str The Table name
        :param record_id: str The Record name
        :param column_name: str The Column name
        :param data: bytes
        :param content_type: str, Default: "application/octet-stream",
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/data/{record_id}/column/{column_name}/file"
        headers = {"content-type": content_type}
        return self.request("PUT", url_path, headers, data=data)

    def delete(
        self, table_name: str, record_id: str, column_name: str, db_name: str = None, branch_name: str = None
    ) -> Response:
        """
        Deletes a file referred in a file column

        Path: /db/{db_branch_name}/tables/{table_name}/data/{record_id}/column/{column_name}/file
        Method: DELETE
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error

        :param table_name: str The Table name
        :param record_id: str The Record name
        :param column_name: str The Column name
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/tables/{table_name}/data/{record_id}/column/{column_name}/file"
        return self.request("DELETE", url_path)

    def transform(self, url: str, operations: dict[str, any]) -> bytes:
        """
        Image transformations
        All possible combinations: https://xata.io/docs/concepts/file-storage#image-transformations

        :param url: str Public or signed URL of the image
        :param operations: dict Image operations

        :return Response
        """
        # valid url ?
        url_parts = url.split("/")
        if 4 < len(url_parts) < 5:
            raise Exception("invalid image url")

        # build operations - turn objects into lists
        ops = []
        for k, v in operations.items():
            if type(v) is dict:
                v = ";".join([str(x) for x in v.values()])
            ops.append(f"{k}={v}")
        # ops = ",".join(["%s=%s" % (k, ";".join(list(v.values()))) if type(v) is dict else f"{k}={v}" for k, v in operations.items()])
        if len(ops) == 0:
            raise Exception("missing image operations")

        region = url_parts[2].split(".")[0]
        file_id = url_parts[-1]
        # signed url
        if len(url_parts) == 5:
            endpoint = "https://%s.xata.sh/transform/%s/file/%s" % (region, ",".join(ops), file_id)
        # public url
        else:
            endpoint = "https://%s.storage.xata.sh/transform/%s/%s" % (region, ",".join(ops), file_id)

        resp = request("GET", endpoint)
        if resp.status_code != 200:
            raise ServerErrorException(f"code: {resp.status_code}, server error: {resp.text}")
        return resp.content



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import logging

from requests import Response, request

from .errors import RateLimitException


class Namespace:
    """
    Parent class for Namespaces
    """

    def __init__(self, client):
        self.client = client
        self.logger = logging.getLogger(self.__class__.__name__)

    def get_scope(self) -> str:
        return self.scope

    def is_control_plane(self) -> bool:
        return self.get_scope() == "core"

    def get_base_url(self) -> str:
        if self.is_control_plane():
            return "https://" + self.client.get_config()["domain_core"]
        # Base URL must be build on the fly as the region & workspace Id can change
        cfg = self.client.get_config()
        return "https://%s.%s.%s" % (cfg["workspaceId"], cfg["region"], cfg["domain_workspace"])

    def request(
        self, http_method: str, url_path: str, headers: dict = {}, payload: dict = None, data: bytes = None
    ) -> Response:
        """
        :param http_method: str
        :param url_path: str
        :headers: dict = {}
        :param payload: dict = None
        :param data: bytes = None

        :return requests.Response

        :raises RateLimitException
        """
        headers = {**headers, **self.client.get_headers()}
        # TODO use "|" when client py min version >= 3.9

        url = "%s/%s" % (self.get_base_url(), url_path.lstrip("/"))
        if payload is None and data is None:
            resp = request(http_method, url, headers=headers)
        elif data is not None:
            resp = request(http_method, url, headers=headers, data=data)
        else:
            resp = request(http_method, url, headers=headers, json=payload)

        # log server message
        if "x-xata-message" in resp.headers:
            self.logger.warn(resp.headers["x-xata-message"])

        if resp.status_code == 429:
            raise RateLimitException(f"Rate limited: {resp.json()}")

        return resp



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from requests import Response

from xata.namespace import Namespace


class Sql(Namespace):

    scope = "workspace"

    def query(
        self,
        statement: str,
        params: list = None,
        consistency: str = "strong",
        db_name: str = None,
        branch_name: str = None,
    ) -> Response:
        """
        This endpoint performs the SQL query across the entire database branch. Set your SQL query in the parameter `query`.

        Path: /db/{db_branch_name}/sql'
        Method: POST
        Response: application/json

        :param statement: str The statement to run
        :param params: dict The query parameters list. default: None
        :param consistency: str The consistency level for this request. default: strong
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :returns ApiResponse
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/sql"
        headers = {"content-type": "application/json"}
        payload = {
            "statement": statement,
            "params": params,
            "consistency": consistency,
        }
        return self.request("POST", url_path, headers, payload)



#
# Licensed to Xatabase, Inc under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Xatabase, Inc licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You
# may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# ------------------------------------------------------- #
# Migrations
# Branch schema migrations and history.
# Specification: workspace:v1.0
# ------------------------------------------------------- #

from requests import Response

from xata.namespace import Namespace


class Migrations(Namespace):

    scope = "workspace"

    def getBranchMigrationHistory(self, payload: dict, db_name: str = None, branch_name: str = None) -> Response:
        """
        Get branch migration history [deprecated]

        Path: /db/{db_branch_name}/migrations
        Method: GET
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error
        Response: application/json

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/migrations"
        headers = {"content-type": "application/json"}
        return self.request("GET", url_path, headers, payload)

    def getBranchMigrationPlan(self, payload: dict, db_name: str = None, branch_name: str = None) -> Response:
        """
        Compute a migration plan from a target schema the branch should be migrated too.

        Path: /db/{db_branch_name}/migrations/plan
        Method: POST
        Response status codes:
        - 200: Example response
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/migrations/plan"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def executeBranchMigrationPlan(self, payload: dict, db_name: str = None, branch_name: str = None) -> Response:
        """
        Apply a migration plan to the branch

        Path: /db/{db_branch_name}/migrations/execute
        Method: POST
        Response status codes:
        - 200: Schema migration response with ID and migration status.
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/migrations/execute"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def getBranchSchemaHistory(self, payload: dict, db_name: str = None, branch_name: str = None) -> Response:
        """
        Query schema history.

        Path: /db/{db_branch_name}/schema/history
        Method: POST
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error
        Response: application/json

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/schema/history"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def compareBranchWithUserSchema(self, payload: dict, db_name: str = None, branch_name: str = None) -> Response:
        """
        Compare branch with user schema.

        Path: /db/{db_branch_name}/schema/compare
        Method: POST
        Response status codes:
        - 200: Schema comparison response.
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/schema/compare"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def compareBranchSchemas(self, payload: dict, db_name: str = None, branch_name: str = None) -> Response:
        """
        Compare branch schemas.

        Path: /db/{db_branch_name}/schema/compare/{branch_name}
        Method: POST
        Response status codes:
        - 200: Schema comparison response.
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/schema/compare/{branch_name}"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def updateBranchSchema(self, payload: dict, db_name: str = None, branch_name: str = None) -> Response:
        """
        Update Branch schema

        Path: /db/{db_branch_name}/schema/update
        Method: POST
        Response status codes:
        - 200: Schema migration response with ID and migration status.
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/schema/update"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def previewBranchSchemaEdit(self, payload: dict, db_name: str = None, branch_name: str = None) -> Response:
        """
        Preview branch schema edits.

        Path: /db/{db_branch_name}/schema/preview
        Method: POST
        Response status codes:
        - 200: OK
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error
        Response: application/json

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/schema/preview"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def applyBranchSchemaEdit(self, payload: dict, db_name: str = None, branch_name: str = None) -> Response:
        """
        Apply edit script.

        Path: /db/{db_branch_name}/schema/apply
        Method: POST
        Response status codes:
        - 200: Schema migration response with ID and migration status.
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/schema/apply"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)

    def pushBranchMigrations(self, payload: dict, db_name: str = None, branch_name: str = None) -> Response:
        """
        The `schema/push` API accepts a list of migrations to be applied to the current branch.  A
        list of applicable migrations can be fetched using the `schema/history` API from another
        branch or database.  The most recent migration must be part of the list or referenced (via
        `parentID`) by the first migration in the list of migrations to be pushed.  Each migration
        in the list has an `id`, `parentID`, and `checksum`. The checksum for migrations are
        generated and verified by xata.  The operation fails if any migration in the list has an
        invalid checksum.

        Path: /db/{db_branch_name}/schema/push
        Method: POST
        Response status codes:
        - 200: Schema migration response with ID and migration status.
        - 400: Bad Request
        - 401: Authentication Error
        - 404: Example response
        - 5XX: Unexpected Error
        - default: Unexpected Error

        :param payload: dict content
        :param db_name: str = None The name of the database to query. Default: database name from the client.
        :param branch_name: str = None The name of the branch to query. Default: branch name from the client.

        :return Response
        """
        db_branch_name = self.client.get_db_branch_name(db_name, branch_name)
        url_path = f"/db/{db_branch_name}/schema/push"
        headers = {"content-type": "application/json"}
        return self.request("POST", url_path, headers, payload)
